﻿#pragma once
#include <QColor>
#define MaxTrackType 10
#define MaxNature 15
#define MaxPoints 7
const int Coli = 5;
const int Peaki = 6;
const int Hikei = 7;
const int Woodi = 13;
const int Wateri = 14;
const int Riveri = 11;
const int Creeki = 12;
const int Raili = 15;
const char tr_give_way = 1;
const char tr_stop = 2;
const char tr_lights = 5;
const char tr_one_way = 10;
const float pf_short = 0.2;
const float pf_flat = 10;
const float pf_time = 10;



//! setting for poweranalysis
typedef struct PowerAnalysisSetting{
    double ws,wd;           //!< Wind strenght ws [bft] and direction wd [°]
    double m;               //!< system weigth [kg]
    double cr;              //!< rolling resistance
    double T;               //!< Temperature at sea level [°C]
    double p0;              //!< air pressure at sea level [hPa]
    double Pvd,Pmax,PT;     //!< Fermi Power function parameter [km/h,W,km/h]
    double cwAvd, cwAmax,cwAmin,cwAT;
                            //!< Fermi CwA function parameter[,,km/h,]
    int fightwind;
    double abrake;          //!< maximum brake accelartion [m/s²]
    double aradial;         //!< maximum radial accelaration [m/s²]
}PowerAnalysisSetting;

/*OSM*/
enum branch {l,r,lefttop,leftbottom};
enum WayColor {none, blue, yellow, green, red, black, pink};
enum readdirection {fw,bw};
enum inputstate {routeshort, routeflat, routemountain,routeclimb,salesman,
                 routeSTRONG,editdata,selectlayer, routetwitch,routetime};
enum climbstate {increasing, decreasing, outofuse, climbstart, climbend, failure};
enum dotstate {crossdot, waydot, nonedot, crossdotmarked};
struct STRONG;
class KOO;
class Way;


//! container struct for Track to hold distance and height data
typedef struct trackpoint{
	float h,d;
}trackpoint;
//! %routing container struct, connects cross via Way with each other
/*! This is one of the most important data types oft BTP. It provides the
    connection from one cross to another cross at the KOO nb. Every Neighbour
    has a brother Neighbour brthr which differs in Neighbour::rd attribute.
    A Neighbour struct contains the heightdata of a road and has a referenze
    tp its KOO, which is stored at the way the Neighbour struct lives on.
    Neighbour structs are generated in the crossdata of a KOO by the WayHeight
    class. Furthermore the Neighbour is also the element of a list linking to
    other Neighbour structs of a KOO, as a cross at least has two Neighbours
    except its the endpoint of a dead end.
*/
typedef struct Neighbour{
    /*! If a routing algorithm reconstructs a Track from a Neighbour array
        than (concerning both the heightdata and the KOO) it is important
        whether treferenced Way of the OSM data was defined forward or backward
    */
    readdirection rd;
    Way *way;							//!< Neighbour lives uses this Way
    KOO *nb;							//!< the KOO, this Neighbour links to
    //! related Neighbour with opposite read direction supposed to store brthr
    Neighbour *brthr;
    float hm;							//!< total height meter, for routing
    float d;							//!< distance
    float t;							//!< time to overcome this road piece
    int nodec;							//!< count of used KOO in way
    int tpc;							//!< count of trackpoint
    /*! traffic restriction used in routing algorithm, the higher the number
     *  the larger the restriction (at the end of the road):
     *   0 - no restrictions
     *   5 - give away
     *  20 - stop
     *  50 - traffic lights
     * 255 - against one way road
    */
    char tr;
    KOO** node;							//!< reference to KOO (of way)
    trackpoint* tp;						//!< heighdata of this Neighbour
    Neighbour* next;					//!< next Neighbour, set to NULL if last
}Neighbour;
//! container struct to hold data of a cross
/*! If a KOO has crossdata, than this KOO is marked as cross. The crossdata
 *  consists of\n
 *  1. a Neighbour list linking this cross to others\n
 *  2. a STRONG struct, where %routing algorithm can operate on. The STRONG
 *     struct allows the algorithms to trace back their way from one STRONG
 *     struct to another and therefor from one cross back to another
*/
typedef struct crossdata{
    Neighbour *neighbours;             //!< neigbours to link to other cross
    STRONG *s;                          //!< memory for %routing algorithms
}crossdata;
//! container struct to hold data of a coordinate
/*! For everx node which is declared in %OSM data one KOO will be constructed
 *  and stored in a AVL tree. If the cd attribute is unequal NULL this KOO
 *  represents a cross
*/
class KOO {
    public:
        KOO();
        KOO(qint64 id);
        qint64 id;                          //!< id inherited from %OSM data
        char balance;                       //!< AVL tree balance
        KOO *l,*r,*prev;                    //!< AVL tree pointers
        double lon,lat;                     //!< the actual coordinate data
        //! state of the KOO when used by different classes
        /*! multiple classes use this attribute to labe the KOO. The following
         *  labels do exist:\n
        *   -1: recently inserted by OSM::insert() function\n
        *   0: unused\n
        *   1: right in the middle in the Way of a Neighbour, no cross\n
        *   2: cross, at the ends of Neighbour\n
        *  	3: element of map (village, town, summit, ...)\n
        *   10: will be save as cross by OSM::save_cross()\n
        *   11: will NOT be save as cross by OSM::save_cross()\n
        *   45: route drag and drop: belongs to track
        *   46: route drag and drop: belongs to favourites\n
        */
        char status;
        crossdata* cd;                      //! holds data if cross
};
//! container for line geodata
/*! A way container contains a KOO array which can be drawn on the map.
    Furthermore the data is neccesary to vreate Neighbour relations in the
    class WayHeight. The type attribute holds the function of the way:\n
    type=-2	relationmember unused\n
    type=-1 raw data    to be classified\n
    type= 0 Motorway\n
    type= 1 Highway\n
    type= 2 primary\n
    type= 3 secondary\n
    type= 4 track\N
    type= 5 track2      (small track, unpaved)\n
    type= 6 wood        (forrest)\n
    type= 7 water\n
    type= 8 river\n
    type= 8 creek\N
    type=98 unuesed raw data to be ignored if saved\n
    type=99 not all KOO of this Way are properly linked\n
*/
class Way{
    public:
        Way();
        Way(qint64 id);
        void init();
        void add_color(WayColor cl);
        qint64 id;                          //!< %OSM way id
        int nodec;                          //!< count of KOO
        KOO** node;                         //!< array KOO pointers
        float minlat,minlon,maxlat,maxlon;  //!< bounding rect, for rendering
        Way *l,*r,*prev;                    //!< AVL tree struct
        short balance;                      //!< AVL tree balance
        char type;                          //!< classification
        char* name;                         //!< name ref owned by lines
        void set_name(char *name);
        //! hiking and cyling mark
        /*! array corresponds to colors: yellow, red, green, blue, black, pink
         *  char value: 1: bar, 2: circle, 3: diagonal, 4: triangle, 5 cross
         */
        char waycolors[6];
        bool IsStair;
        bool IsTunnel;
        bool IsBridge;
        bool IsOneWay;
        bool IsPriorityRoad;
};
typedef struct OSM_Data{
	KOO** nodes;
    Way** waysavl;
} OSM_Data;
//! osm data container list for point like data (summits, towns, ...)
class points{
public:
    points();
    points(KOO* KOO,char* name, float index);
    points(KOO* KOO,char* name, int index);
    points(KOO* KOO,char* name, char index);
    points(points* prev);
    void insert(points* prev);
    points* insert_order_increasing(points* src);
    void remove();
    points* root();
    KOO* k;                             //!< reference to KOO
    float index;						//!< for summit ranking
    //! hike map element
    /*! specifies imporant hiking map elements, small numbers painted first\n
    *   0:  nichts\n
    *   10: Vorfahrtsschild\n
    *   11: Stoppschild\n
    *   12: Ampeln\n
    *   21: Parkplatz\n
    *   22: Haltestelle\n
    *   30: Tankstelle\n
    *   31: Supermarkt\n
    *   32: Bäcker\n
    *   35: Baum\n
    *   38: Schutzhütte\n
    *   39: Informationsschild\n
    *   40: Karte\n
    *   42: Cafe\n
    *   43: Kneipe\n
    *   44: Denkmal\n
    *   44: Museum\n
    *   45: Burg\n
    *   47: Turm\n
    *   50: Restaurant\n
    *   51: Kiosk\n
    *   52: Kirche\n
    *   53: Ruine\n
    *   57: Windrad\n
    *   60: Aussichtspunkt\n
    *   61: Höhle\n
    *   62: Schranke\n
    *   64: Wegweise\n
    *   70: Quelle\n
    *   71: Wasserfall\n
    */
    char* name;                         //!< %OSM name of point
    points* next, *prev;};
//! osm data container list for line like data (roads, rivers)
typedef struct lines{
    Way* way;                           //!< reference element of this list
	lines* next;
    char* name;                         //!< %OSM name of line
}lines;
//! osm data container list for relation data (multiple lakes and forrests)
class relation{
    public:
        relation();
        Way** members;
        int memberscount;
        char type;
        relation* next;
        char* name;                         //!< %OSM name of relation
        WayColor color;
};


//! main container list climbanalyse
/*! this struct is considered to hold a tree struct of all climbs belonging to
    the same starting point. Consequently the list should only be read back-
    wards as forward it branches.
*/
typedef struct trackpointlist{
	trackpoint t;
	trackpointlist *prev,*next;
}trackpointlist;
//! main container for Track, WayHeight and ClimbAnalyse
/*! This list hold coordinates to define track on map. Used in Track, the list
 *  pointers are ignored as list flexibility is not required. Used in
 *  ClimbAnalyse this list are usually branched to a whole tree of climbs with
 *  common start point.
*/
typedef struct KOOlist{
    KOO* k;                             //!< reference to KOO data
    KOOlist *prev,*next;
    //! current distance, synchronized to trackpoint array
    float d;
    //! labels elements for drag-drop-routing:
	short marked;
    //! index of related element in synchronized trackpoint array
    int tpi;
}KOOlist;

/*MAP*/
//! setting of the light to render the DEM
typedef struct SunSetting{
    float azimut,elevation;
    float I;                            //!< intensity
    float sphi;                         //!< precalculated sin
    float cphi;                         //!< precalculated cos
    float stheta;                       //!< precalculated sin
    float ctheta;                       //!< precalculated cos
}SunSetting;

/*Profil*/
//! setting for 2D plot
typedef struct ProfilSettings{
    QRgb bgcolor,fontcolor,latticecolor;
    bool drawybarlabel,drawxbarlabel,drawtitle,
        drawxlattice,drawylattice;
    int fsnormal, fstitle;				//Schriftgroessen;
    char font[100];
    double xmax,xmin,ymax,ymin;
    int margin;
    int width, height;
    bool autoscale;
    bool singlecolor;
    QRgb fillcolor;
    bool showline;
    QRgb linecolor;
    //! average data over multiple screen pixels to sensfully show fill color
    double datacount;
}ProfilSettings;

/*Profil*/
typedef struct MapSettings{
    double minlat,minlon,maxlat,maxlon;	/*!< map view*/
    QRgb roadcolor[MaxTrackType+1];     //!< color for all different road types
    float roadwidth[MaxTrackType+1];    //!< width for all different road types
    float road_dash_line[MaxTrackType+1];//!< dash line length
    float road_dash_space[MaxTrackType+1];//!< dash line length
    QRgb trackcolor;                    //!< color for track
    float trackwidth;                   //!< width for track
    QRgb summitcolor;
    float summitsize;
    float pointsize[6];                 //!< fontsize of town, village, ...
    QRgb pointcolor[6];                 //!< fontcolor of town, village, ...
    bool drawcontour;
    bool drawsummit;
    bool drawpoints;
    bool drawwood;
    bool drawstreetcontour;
    QRgb woodcolor;
    bool drawater;
    QRgb watercolor;
    bool drawscale;
    bool drawstreetlabel;
    QRgb heightcolor[5];                //!< 5 step color scale
    double heights[5];                  //!< heights of 5 step color scale
    double climblayershading;
    SunSetting sun;                     //!< lighting of relief
    bool drawHikeBike;
    float maxintersection;              //!< for printing labels on map
    float globalzoom;                   //!< controls MAP::zoomsize
}MapSettings;

/*Routing*/
//! main struct for routing purpose
/*! This struct is part of the crossdata of every KOO, which is a cross. In
 *  general, only one struct per cross is needed (shortes or flatest routing),
 *  but to allow drag and drop routing, two STRONG are kept ready at least.
 *  For %STRONG routing algorithms more STRONG structs will be allocated,
 *  depending on the step size.\n
 *  A STRONG struct knows the STRONG which was reached before which is stored
 *  as a void pointer in from. Alternativly, if STRONG struct is still active,
 *  it links to the data which is handeled in a fibunacciheap. The via
 *  attribute references the Neighbour, by which the current STRONG was
 *  reached.\n
 *  As a STRONGcalc is memory intense and this datatype requires almost all
 *  memory it is kept small without additional information as distance or Hm,
 *  this must be reconstructed or is stored in the FibunacciHeap.
 */
typedef struct STRONG{
    //!< pointer to STRONG or FibunacciHeap
    void* from;
    Neighbour* via;						//!< connection to reach this STRONG
}STRONG;
//! FibunacciHeap element to store STRONG data
class STRONGfib{
public:
    float d;							//!< current distance at STRONG
    float hm;							//!< current Hm at STRONG
    STRONG* from;						//!< pointer to related STRONG struct
    KOO* k;								//!< current KOO
    Neighbour* via;						//!< connection to STRONG
};
//! FibunacciHeap element to store STRONG data with climb restrictions
class STRONGPlusfib : public STRONGfib{
public:
    //! Hm ascenting since last big descent section
    float hmup;
    //! Hm descenting since last big descenting section
    /*! %STRONG algorithm will stop, if hmdown/hmup exceeds a userdefined limit
     */
    float hmdown;
};
//! FibunacciHeap element to store STRONG data with climb analysis
class STRONGClimbfib : public STRONGPlusfib{
public:
    float Psum, dP;
};
typedef struct STRONGsetting{
    //! Area restricitions of STRONG routing
    /*!
      restricts the area to calculate to a circle with the radius
      (airline distance * fac + abs)
    */
    double rfac, rabs;
    //! Additional criteria for STRONGPlus defining climb
    int hmstep;                         //!< Algorithm step size
    int hmmax;                          //!< finalization criterium absolute
    double hmperkm;                     //!< hmmax=max(hmmax,radius*hmperkm)
    short dhmin;                        //!< defines climb: min. Hm
    float downtoupmax;                  //!< defines climb: allowed down Hm
    float minP;                         //!< defines climb: min. difficulty
    bool noSTRONGPlus;                  //!< dont apply climb restrictions
    bool ClimbInsteadHm;                //!< maximizes difficulty instead Hm
}STRONGsetting;
//! container for user settings to run ClimbAnalyse
typedef struct ClimbAnalyseSettings{
    double minP;						//!< climb analyse dialog
    double minS;						//!< climb analyse dialog
    double maxdwnHM;					//!< climb analyse dialog
    double minHM;						//!< climb analyse dialog
    bool html;      					//!< climb analyse dialog
    bool map;							//!< climb analyse dialog
    bool list;          				//!< climb analyse dialog
    char ProjectFolder[260];            //!< climb analyse dialog
    char ProjectName[260];              //!< climb analyse dialog
    double mapsize;					    //!< climb analyse dialog
    bool Climblayer;					//!< climb analyse dialog
}ClimbAnalyseSettings;
//! container for user settings to manually generate map
typedef struct ExtMapSettings{
    double minlat,minlon,maxlat,maxlon;	//!< borders
    double mapsize;                		//!< picturesize in megapixel
    bool drawCL;             			//!< draw climblayer
    bool drawSL;     					//!< draw stronglayer
    bool drawct;                        //!< draw current track
}ExtMapSettings;
//! main user setting container
typedef struct config {
    MapSettings MS;
    char mintype,maxtype;				//!< way restrictions
    STRONGsetting mconfig;              //!< STRONGAlgo setting, fast mode
    STRONGsetting sconfig;              //!< STRONGAlgo setting, detail mode
    STRONGsetting cconfig;              //!< STRONGAlgo setting, climb mode
    char SRTM3folder[260];              //!< SRTM3 data folder
    char SRTM1folder[260];              //!< SRTM1 data folder
    char SRTM066folder[260];            //!< SRTM066 data folder
    char SRTM018folder[260];            //!< SRTM018 data folder
    char DEM_level;                     //!< to remember last used DEM
    char start_btp[260];                //!< osm data to load at beginning
    inputstate input;					//!< current routing mode
    int drawSTRONGlayer;				//!< create mapdialog setting
    int drawRoute;						//!< create mapdialog setting
    int drawClimblayer;                 //!< create mapdialog setting
    ClimbAnalyseSettings CA;
    ProfilSettings PS;
    PowerAnalysisSetting PA;
    int plotspeed;                      //!< Plotting power calc results
    int plotpower;                      //!< Plotting power calc results
    int plotbpm;                        //!< Plotting bpm data

}config;
//! base information to draw road section in STRONGlayers
typedef struct STRONGlayerele{
    Neighbour* n;                       //! connection to be drawn
    float index;                        //! pen width is calculated from index
}STRONGlayerele;
//! list elements to handle multiple STRONGlayers
typedef struct STRONGlayer{
    STRONGlayerele* data;
    int ncount;                         //! count of indexed road sections
	STRONGlayer* prev;
	STRONGlayer* next;
}STRONGlayer;

/*WayHeight*/
typedef struct PortionInstruction{
	double cos;
	trackpoint *dest;
}PortionInstruction;
typedef struct PortionInstructionCol{
    Neighbour* nb;
	PortionInstruction* PI;
	double SumCos;
	int Pcount;
	trackpoint *src;
	PortionInstructionCol* next;
}PortionInstructionCol;

/*ClimbAnalyse*/
typedef struct nbpoolele{
	int mintp;
    Neighbour* nb;
}nbpoolele;
typedef struct crosspoolele{
	float h;
	KOO* k;
}crosspoolele;
//! container struct for ClimbAnalyse
class climb{
public:
    climb();
    climb(climb *c1, climb *c2);
    void init(double d, double Hm);
    void init();
    void clone(climb* c);
    climbstate state;                   //!< current state at ClimbAnalyse
    short branches;                     //!< count of branching chilf climbs
    KOO* home;                          //!< cross, climb refers to
    trackpointlist* t;                  //!< track profil so far, to trace back
    trackpointlist* tst;                //!< start point in track profil
    trackpointlist* tend;               //!< end point in track profil
    //!@{
    //! climb properties
	float P, Pmax, Prec, d, hend, hstart, hmdown,hmup;
    //!@}
    climb * next,*prev;
    climb *cst;                         //!< start point as climb struct
    climb *root;                        //!< root of the whole climb tree
    Neighbour* via;                     //!< last used link
};
typedef struct CAfib{
	float h;							//Distanz der KOO
	climb* c;
}CAfib;
class Cfib : public STRONGfib{
public:
    float hmup,hmdown,P;
};								//Hilfstruktur fuer ClimbDijkstra

typedef struct climblayerelement{
	double **dharray;
	int dhcount;
	double **KOOarray;
	int KOOcount;
    char* name;
    //!@{
    //! climb properties
	float P,d, hend, hstart, hmdown,hmup, slope,bend,LocRank;
    //!@}
    double midlat,midlon;
}climblayerelement;
typedef struct climblayer{
	climblayerelement* climb;
	long climbcount;
	climblayer* prev;
	climblayer* next;
}climblayer;
typedef struct templatepoint{
    int x,y;
    int dx,dy;
    short rprev;
    short rnext;
}templatepoint;
