#include "DataTyps.h"

Way::Way(){
    init();
}
Way::Way(qint64 id){
    init();
    this->id = id;
}
void Way::init(){
    id = -1;
    nodec = 0;
    node = NULL;
    minlat = 90;
    minlon = 180;
    maxlat = -90;
    maxlon = -180;
    l = r = prev = NULL;
    balance = 0;
    type = -1;
    name = NULL;
    waycolors[0] = waycolors[1] = waycolors[2] = waycolors[3] = waycolors[4] = waycolors[5] = 0;
    IsStair = IsTunnel = IsBridge = IsOneWay = IsPriorityRoad = false;
}

void Way::add_color(WayColor cl){
    switch(cl){
        case yellow: this->waycolors[0] = true; break;
        case pink: this->waycolors[1] = true; break;
        case red: this->waycolors[2] = true; break;
        case green: this->waycolors[3] = true; break;
        case blue: this->waycolors[4] = true; break;
        case black: this->waycolors[5] = true; break;
    }
}
void Way::set_name(char *name){
    if(name != NULL){
        this->name = new char[strlen(name)+1];
        strcpy(this->name,name);
    }
}

relation::relation(){
    name = NULL;
    color = none;
    members = NULL;
    memberscount = 0;
    type = -1;
    next = NULL;
}
KOO::KOO(qint64 id){
    l = r = prev = NULL;
    lat = 200;
    lon = 200;
    balance = 0;
    status = -1;
    cd = NULL;
    this->id = id;
}

KOO::KOO(){
    l = r = prev = NULL;
    lat = 200;
    lon = 200;
    balance = 0;
    status = -1;
    cd = NULL;
    id = -1;
}
points::points(){
    k = NULL;
    index = 0;
    name = NULL;
    next = prev = NULL;
}
points::points(KOO *koo, char *name, float index){
    k = koo;
    this->index = index;
    if(name != NULL){
        this->name = new char[strlen(name)+1];
        strcpy(this->name,name);
    }
    else
        this->name = NULL;
    next = prev = NULL;
}
points::points(KOO *koo, char *name, int index){
    k = koo;
    this->index = float(index);
    if(name != NULL){
        this->name = new char[strlen(name)+1];
        strcpy(this->name,name);
    }
    else
        this->name = NULL;
    next = prev = NULL;
}
points::points(KOO *koo, char *name, char index){
    k = koo;
    this->index = float(index);
    if(name != NULL){
        this->name = new char[strlen(name)+1];
        strcpy(this->name,name);
    }
    else
        this->name = NULL;
    next = prev = NULL;
}

points::points(points *prev){
    k = NULL;
    index = 0;
    name = NULL;
    insert(prev);
}
points* points::root(){
    points* root = this;
    while(root->prev != NULL)
        root = root->prev;
    return root;
}
points* points::insert_order_increasing(points* src){
    points *t = this;
    if(t->index <= src->index){
        while(t->prev != NULL && t->prev->index < src->index)
            t = t->prev;
        src->prev = t->prev;
        src->next = t;
        t->prev = src;
        if(src->prev != NULL)
            src->prev->next = src;
        return src;
    }
    else{
        while(t->next != NULL && t->next->index > src->index)
            t = t->next;
        src->next = t->next;
        src->prev = t;
        t->next = src;
        if(src->next != NULL)
            src->next->prev = src;
        return this;
    }
}

void points::insert(points* prev){
    if(prev != NULL){
        this->prev = prev;
        this->next = prev->next;
        prev->next = this;
        if(next != NULL)
            next->prev = this;
    }
    else{
        next = NULL;
    }
}
void points::remove(){
    if(prev != NULL)
        prev->next = next;
    if(next != NULL)
        next->prev = prev;
}
climb::climb(){
    init();
}
void climb::init(){
    state = outofuse;
    branches = 0;
    home = NULL;
    t = tst = tend = NULL;
    P = Pmax = Prec = d = hend = hstart = hmdown = hmup = 0;
    next = prev = cst = root = NULL;
    via = 0;
}
climb::climb(climb *c1, climb *c2){
    if(c2 == NULL && c1 == NULL){
        init();
        return;
    }
    else{
        if(c2 == NULL){
            clone(c1);
            return;
        }
        if(c1 == NULL){
            clone(c2);
            return;
        }
        state = outofuse;
        branches = c1->branches + c2->branches;
        home = NULL;
        d = c1->d + c2->d;
        hmup = c1->hmup + c2->hmup;
        hmdown = c1->hmdown + c2->hmdown;
        hstart = c1->hstart;
        hend = c2->hend;
        Pmax = Prec = 0;
        next = c2->next;
        prev = c1->prev;
        cst = root = NULL;
        via = 0;
        // try to determine climb attribute
        if( hmup > 70 &&
            hmdown / hmup < 0.1 &&
            hmup*hmup/ d > 5 &&
            hmup          / d > 0.05){
            branches = 1;
            P = hmup*hmup/ d;
        }
        // check if subclimbs are harder than merged climb
        if(P < c1->P || P < c2->P){
            P = c1->P + c2->P;
            branches = c1->branches + c2->branches;
        }
    }
}
void climb::clone(climb* c){
     state = c->state;
     branches = c->branches;
     home = c->home;
     t = c->t;
     tst = c->tst;
     tend = c->tend;
     P = c->P;
     Pmax = c->Pmax;
     Prec = c->Prec;
     d = c->d;
     hend = c->hend;
     hstart = c->hstart;
     hmdown = c->hmdown;
     hmup = c->hmup;
     next = c->next;
     prev = c->prev;
     cst = c->cst;
     root = c->root;
     via = c->via;
}

void climb::init(double d, double Hm){
    this->d = d;
    if(Hm > 0){
        hmup = Hm;
        hmdown = 0;
    }
    else{
        hmup = 0;
        hmdown = -Hm;
    }
    if( hmup > 70 &&
        hmdown / hmup < 0.1 &&
        hmup*hmup/ d > 5 &&
        hmup          / d > 0.05){
        branches = 1;
        P = hmup*hmup/ d;
    }
    else{
        P = 0;
        branches = 0;
    }
}
