﻿/*Liest Rohdaten aus OSM-Datein ein*/

#pragma once
#include <string>
#include <stdio.h>
#include <cstdlib>
#include <QtWidgets/QMessageBox>
#include "DataTyps.h"
#include "MathExt.h"
#include "HeightData.h"

using namespace std;
//! BTP3 database, created from OpenStreetMap data
/*! This class manages all data content of BTP3. Map rendering as well as
 *  routing algorithm directly access this data.
 *  The database consists of two AVL trees containing $osm nodes and $osm ways.
 *  In BTP3 language this means KOO and Way. Main content of a Way is a KOO
 *  reference array. The third big $osm data type - the Relation, a collection
 *  of ways - does not eyist in osm class.
 *  For Rendering purpose there exist several list classifying the KOO and
 *  Way, these lists are points and lines respectivly. The following lines
 *  exist:\n
 *  motorWay, highway, primary, secondary, track, track2, river, creek,
 *  wood, water, NoHeight\n
 *  The following points exist:\n
 *  metropolis, city, town,	village, locality, col, peak;\n
 *  The classification is done by nodehandling and wayhandling.
 *
 *  This data is generated from OpenStreetMap XML files *.osm. Within this
 *  class this kind of file is read by read_OSM.
 *
 *  HeighData of road like lines is calculated by the seperate class WayHeight.
*/
class OSM{
	public:
		OSM(string filename);
		OSM();
		~OSM();
        void read_OSM();				//!< read osm file
        void get_KOO_Waysavl();			//!< links node IDs to nodes recursivly
        void check_lines();				//!< delete Way with unresolved nodes
        void trim_points();             //!< reduces number of hike map elements
		void setfilename(string fileName);
        Way* get_first_Way();           //!< returns root of Way-AVL tree
        KOO* get_first_KOO();           //!< returns root of node-AVL tree
        Neighbour* add_Neighbour(KOO* k, KOO** node, KOO* nb, Way* w,
								float d, int KOOc, readdirection rd);
                                        //!< add Neighbour cross to k
        void initiallists();			//!< initial class's containers
        void RelationstoFace(relation** l,lines** f);
                                        //!< collect all Way to one relation
        void RelationstoWay(relation** l);
                                        //!< pass realtion attributes to way
		void reset_KOO_status(KOO* k, char i);
                                        //!< reset KOO status to i
        void restore_KOO_status();      //!< lines --> 1, points --> 3
        //! summit ranking uses t-functions of heighdata
        void calc_peak_index_pro(HeightData* h);

        void sort_peaks();				//!< primitive sort algorithm


        //*Bounding rect
        void calc_brect();              //!< refresh bounding rect of dataset
        double getminlat();             //!< return lsat calced bounding rect
        double getminlon();             //!< return lsat calced bounding rect
        double getmaxlat();             //!< return lsat calced bounding rect
        double getmaxlon();             //!< return lsat calced bounding rect

        // container data
        //!@{
        //! container data
        short mintype,maxtype;
        //! array containing line like data elements
        /*! index 0: türkies, Autobahn
         *  index 1: rot, Bundesstraße
         *  index 2: orange, Staatsstraße
         *  index 3: gelb, Kreisstraße
         *  index 4: weiß, kleine, asphaltierte Straßen und Wege
         *  index 5: schwarz dick, Forstautobahnen
         *  index 6: gestrichelt, Wald/Forstwege mit noch fahrbaren Untergrund
         *  index 7: punkt-strich, Wald/Forstwege mit schlechtem Untergrund
         *  index 8: schwarz dünn, Pfade mit glattem Untergrund
         *  index 9: gepunktet, Pfade
         *  index 10: gedreieckt, Pfade mit steinigen Untergrund
         *  index 11: Wald
         *  index 12: Wasserflächen
         *  index 13: Flüsse
         *  index 14: Bäche
         */
        lines** ln[MaxNature+1];          //!< tracks
        //**motorWay,**highway,**primary,**secondary,**track,**track2,
        //	  **river,**creek,** wood,**water, **NoHeight;
        //! array containing point like data elements
        /*! index 0:        100000 : tag city
         *  index 1:         10000 :
         *  index 2:          3000 : tag town
         *  index 3:            20 : tag village, suburban
         *  index 4:            20 : tag hamlet, locality, isolated_dwelling
         *  index 5: col, mountain pass
         *  index 6: peak,summit
         *  index 7: hike map element
        */
        points** pt[MaxPoints+1];
        relation **waterr,**woodr, **router;//!@}

        //!@{
        //! check whether data lies within defined rectangular area
		short isin(lines* l,double minlat, double maxlat, 
							double minlon, double maxlon);
		short isin(points* l,double minlat, double maxlat, 
							 double minlon, double maxlon);
		short isin(KOO* k,double minlat, double maxlat, 
						  double minlon, double maxlon);
        short isin(Way* w,double minlat, double maxlat,
						  double minlon, double maxlon);
        short iscompletein(Way* w,double minlat, double maxlat,
								  double minlon, double maxlon);
        //!@}

        //aktuelle Status
        int get_status();               //!< returns current class state
        int count_cross(KOO *k);        //!< counts crossdata recursivly
        int count_STRONG(KOO *k);       //!< counts STRONG data recursivly

		/*Berechnete Daten Speichern und laden*/
        //! save dataset as btp
		void save(char* filename, double minlat, double maxlat, 
				  double minlon, double maxlon);
        //! load btp to dataset
		void load(char* filename);
		
		/*Daten editieren*/
        bool insert_cross(Way* w, double latc, double lonc);
        void add_user_track(KOOlist *kl, trackpoint *tpl, int tc);
		void select_cross(double lat, double lon,
						  KOO** kdest, float* d,KOO* ksrc);
        void select_cross(double lat, double lon, char mintype, char maxtype,
                          KOO** kdest, float* d,KOO* ksrc);
        Way* select_Way(double lat, double lon);
        void change_Waytype(Way* w, char newtype);
        void del_Way(Way* w);
        void normalize_Hm(KOO* k);      //!< removes total height differences from Neighbour relation
        void invert_Hm(KOO* k);         //!< switches descenting and ascenting vertical meters in Neighbour relation
        void reset_Hm(KOO* k);          //!< resets Neighbour Hm to ascenting vertical meters
        void points_Hm(KOO* k);         //!< accumulates section difficulties instead of vertical meters in Neighbour relation
        void swap_Hm_d(KOO* k);         //!< swapes distance and height meters
        float d(KOO* k1, KOO* k2);      //!< calculates distance between two coordinates
        //! recaculates Neighbour time with give function
        void refresh_time(KOO* k, float (*t)(Neighbour* n, void* inst), void* buf);
        KOO* search(qint64 id, KOO* s);	//!< returns KOO with ID starting at branch s
        //! solves conflictiv KOO and crossdata when loading several btp datasets
		void merge(KOO* knew, KOO* kold);
        //! delete all data lying within the rectangular area
		void deleteRect(double minlat, double maxlat, 
						double minlon, double maxlon);
        void info();
        void assign_traffic_restrictions();
        void assign_traffic_restrictions(KOO* k);
	private: 
        short check_AVL(Way* w);        //!< check AVL tree construction
        //! search for tag a in string s and extract to qint64
        qint64 extract_valuei(char* s, const char* a);
        //! search for tag a in string s and extract to double
        double extract_valuef(char* s, const char* a);
        //! search for tag a in string s and returns it
        char* extract_valuec(char* s, const char* a);
        //! extrac bounding box specified in osm file
        void extract_bbox(char* s);
        void quotationreplace(char* a);	//!< replaces " by '
        void htmlreplace(char* c);		//!< replaces &#39; and &#34;
        double min_distance(Way* wc, double lat, double lon,
                            int* mini, double *dti);
        //! osm usage status variable
        /*!
         *  status 0: nothing happened but osm class instanciated
         *  status 1: $osm file name is determined
         *  status 2: $osm file was read
         *  status 3: KOO of Way are resolved, i.g. get_KOO_Waysavl() applied
         */
        int status;
        string filename;                /*!< $osm file name*/

		/*Hauptdatenstruktur*/
        OSM_Data D;                     /*!< main data pointer*/
        qint64 avail_ID_max;            /*!< free ID for user input */
        //! AVL tree handling: naiv insert of existing KOO into data subtree s
        KOO* append(KOO* k);
        //! AVL tree handling: naiv insert of id into data subtree s
        KOO* insert(qint64 id, KOO* s);
        //! AVL tree handling: subroutine of insert(), returns root element
        KOO* rotate(KOO** s, branch b);
        //! AVL tree handling: accurate insert of id into data subtree s
        KOO* insert_AVL(qint64 id, KOO** s);

        //! AVL tree handling: naiv insert of id into data subtree s
        Way* insert(qint64 id, Way* s);
        //! AVL tree handling: subroutine of insert(), returns root element
        Way* rotate(Way** s, branch b);
        //! AVL tree handling: accurate insert of id into data subtree s
        Way* insert_AVL(qint64 id, Way** s);		//einfgn. AVL-Baum-konform
        //! AVL tree handling: search for id and return related Way
        Way* search(qint64 id, Way* s);
        //! resolve KOO id to KOO pointer
        void get_KOO_Waysavl(Way* l,short next);
        //! check recursivly completeness of KOO pointers
        /*! recursivly checks whether all KOO id are have been succesfully
         *  transformed into a KOO pointer, if not the concerning lines is
         *  deleted.
         */
        void check_lines_rek(lines** l);
        void way_to_node(Way *l);
        void delete_AVL(Way* l);
        void delete_AVL(KOO* l);
        void delete_lines(lines* l);
        void delete_points(points* l);
        void delete_relation(relation* l);
        void alloc_crossdata(KOO* k);
        void calc_brect(Way* l);

        //! $osm-file reading: dealing with node data lines in file
		void nodehandling(FILE *f,char* line,char* lineattr,char* linename);
        //! $osm-file reading: dealing with way data lines in file
        void Wayhandling(FILE *f,char* line,char* lineattr,char* linename,
                         qint64* currentseek);
        //! $osm-file reading: dealing with relation data lines in file
		void relationhandling(FILE *f,char* line,char* lineattr,char* linename,
                         qint64* currentseek);
        //! try to find hiking map characteristic attributes in string line
        void check_hike_type(char* line, float* hiketype);
        //! add a Way to a OSM_Data liones element with a given name
        void add_line(lines** l,Way* Way,char* name);
        //! add a KOO to a points data element, decreasing index order
        points* add_points(points** dest,points* src);
        //! add a relation to a OSM_Data relation element with a given name and index
        void add_relation(relation** l, relation* r);
        //! invert elements of a list
        void invert_points(points** l);
        //! remove to much natural=tree elements
        void trim_nodes(float index, double mind, bool demand_names);
        void trim_nodes(float index, int dgrid, bool demand_names,
                        QList<points *> **elegrid, int cx, int cy);
        void trim_names(int dgrid,
                        QList<points *> **elegrid, int cx, int cy);
        //! bounding rect of the osm data set
        double minlat,maxlat,minlon,maxlon;
        //! claimed bounding rect of the &osm data file
        double dminlat,dmaxlat,dminlon,dmaxlon;
        //! refresh bounding rect regarding points data
        void calc_brect(points** l);
        //! refresh bounding rect regarding lines data
        void calc_brect(lines** l);

        //!@{
        //! save btp data to btp file
        /*! save data, at base function check wheter element lies within the
         *  bounding rect (minlon,minlat,maxlon,maxlat), eventually count the
         *  elements which were finally save with c pointer
         */
        void save_node(KOO* k, FILE* f);
        void save_Way(Way* w, FILE* f);
        void save_cross(KOO* k, FILE* f, int* c, double minlat,
                        double maxlat, double minlon, double maxlon);
        void save_lines(lines** l, FILE* f, double minlat, double maxlat,
                        double minlon, double maxlon);
        void save_points(points** p, FILE* f, double minlat, double maxlat,
                         double minlon, double maxlon);
        //!@}

        //!@{
        //! load btp data from btp file
		void reset_KOO_status(KOO* k);
		KOO* load_node(FILE* f);
        Way* load_Way(FILE* f);
		void load_cross(FILE* f);
        void load_lines(lines** l, FILE* f);
        void load_points(points** p, FILE* f);
        void replace_Way_KOO(Way* w, KOO** newKOO, int newlength);
        //!@}

        //! find common cross height if Neighbour relation indicates different
        void equal_crossheight(KOO* k);

        //!@{
        //! data set editting functions
        void extract_Way(lines* l, Way* w, Way* k, char mintype, char maxtype);
        void del_lines(Way* w, lines** l, char* keepname);
        void del_Neighbour(KOO* k, Way* w);
        void del_Neighbour(KOO* k, Neighbour* n);
        void del_Way(Way* w, double minlat, double maxlat,
					 double minlon, double maxlon);
        //!@}
};

