﻿#ifndef PLOT2D_H
#define PLOT2D_H
#include <math.h>
#include "DataTyps.h"
#include <QPixmap>
#include <QPainter>
#include <QRect>
#include <QBrush>
#include <QMouseEvent>
#include <QString>
#include <QFontMetricsF>
#include <QClipboard>
#include <QtWidgets/QMessageBox>
#include <QApplication>

//! plotting 2D-Data sets as the height profile of a track or poweranalysis data
/*! This plotting class is based on "C++ GUI Programmierung mit Qt 4:
 *  Die offizielle Einführung (Programmer's Choice)". It allows plotting profil
 *  data set und two line data sets.
*/
class plot2D : public QWidget{
    Q_OBJECT
public:
    plot2D(QWidget *parent = 0);
    plot2D(plot2D *src);                //!< deep copy from existing plot
    ~plot2D();
    //! setting one dataset and repaint plot
    void set_data(double** dataxy, int count);
    //! setting three datasets (1=points, 2/3=lines) and repaint plot
    void set_data(double** data1xy, int count1,
                  double** data2xy, int count2,
                  double** data3xy, int count3);
    void set_labels(QString title, QString xbarlabel, QString ybarlabel);
    QSize sizeHint() const;
    QSize minimumSizeHint() const;
    void autoset_axis();                //!< autoscale plot to full data view
    void set_marker(double x);          //!< draws vertical black line at x pos
    double get_xmin();
    double get_xmax();
    void set_size(int w,int h);
    void save_to_png(QString filename);
    void settings(ProfilSettings* s);   //!< asign settings to plot

public slots:
    void zoom_in();
    void zoom_out();
    void refresh_pixmap();              //!< repaints the plot's pixmap
    void data_to_clipboard();
private:
    double *xdata, *ydata;              //!< profile data
    double *xdata1, *ydata1;            //!< linedata red
    double *xdata2, *ydata2;            //!< linedata blue
    int count, count1, count2;          //!< data point count
    QPixmap pm;                         //!< plot pixel data
    ProfilSettings s0;                  //!< intrinsic settings
    ProfilSettings* s;                  //!< assigned settings to be used
    int xsteps, ysteps, showzoombox, xold, yold;
    double xstep, ystep, xstep0, ystep0;//Achsskalen
    double xprecission, yprecission;    //!< axis label precission
    double linewidth;
    double marker;
    float textheight, titleheight;      //!< text size
    QPoint mouseposition;
    QString title, xbarlabel, ybarlabel;
    QRect dataarea;                     //!< coordinates of plot inset
    QRect zoombox;                      //!< rect to be drawn green when zoomed
    double xtoxdata(int x);             //!< transform pic-px to data value
    double ytoydata(int y);             //!< transform pic-px to data value
    double xdatatox(double x);          //!< transform data value to pic-px
    double ydatatoy(double y);          //!< transform data value to pic-px
    void add_axis_labels(QPainter* p);  //!< rendering subfunction
    void add_grid(QPainter* p);         //!< rendering subfunction
    void add_data(QPainter* p);         //!< rendering subfunction
    QColor slopetocolor(double x);      //!< calc fill color of profil data
    void init_s0();                     //!< initialize intrinsic settings
signals:
    void keypressed(QKeyEvent *event);
    void mousex(double x);              //!< mouse's x value (data, not pixels)
    //! indicate wish to be repainted
    void call_zoom_repaint(double xmin, double xmax);
    void call_fullview();               //!< indicate wish to be rescaled
    void clicked(double x);
protected:
    void paintEvent(QPaintEvent *);
    void resizeEvent(QResizeEvent *);
    void mousePressEvent(QMouseEvent *event);
    void mouseMoveEvent(QMouseEvent *event);
    void mouseReleaseEvent(QMouseEvent *event);
    void keyPressEvent(QKeyEvent *event);

};

#endif // GRAPH_H
