#include "tsp.h"

TSP::TSP(double** data,int c){
    this->data = data;
    this->c = c;

}
edge* TSP::BranchAndBound_1way(double* distance){
    edge* solution = BranchAndBound(distance,1);
    return solution->start();
}
edge* TSP::BranchAndBound(double* distance){
    return sort(BranchAndBound(distance,0),0);
}
edge* TSP::BranchAndBound(double* distance, int maxi){
    // calculate maximum heap size
    double maxHeap = 1.0*free_RAM_Bytes()/( c*c*sizeof(double)
                                            +1*sizeof(TSP)
                                            +c*sizeof(edge));
    FibHeap f;
    TspMatrix* t = new TspMatrix(data,c);
    TspMatrix* t2;
    TspMatrix* solution = t->clone();
    bool estimated = false;
    f.ins(float(t->lower_bound()),t);
    solution->primitive_solution();
    t = (TspMatrix*) f.extmin();
    int fmax = 0;
    while(t != NULL && t->lower_bound() < solution->lower_bound()){
        // avoid heap overflow, cut off to setimated solution instead
        fmax = qMax(fmax,f.size());
        if(f.size() > maxHeap){
            estimated = true;
            while(f.size() > 2){
                delete t;
                t = (TspMatrix*) f.extmin();
            }
        }
        // there was a valid branch handle it
        if(t->finished(maxi)){
            if(t->lower_bound()<solution->lower_bound())
                solution = t;
            else
                delete t;
            t = (TspMatrix*) f.extmin();
        }
        else{
            double lb_before = t->lower_bound();
            t->reduce();
            double lb_after = t->lower_bound();
            t2 = t->clone();
            // handle accept branch
            t->accept_candidate();
            if(lb_before < lb_after){
                f.ins(t->lower_bound(),t);
                t = (TspMatrix*) f.extmin();
            }
            // handel refuse branch
            double d_future = t2->refuse_candidate();
            if(d_future < solution->lower_bound())
                f.ins(d_future,t2);
            else
                delete t2;
        }
   }
    while (t != NULL) {
        delete t;
        t = (TspMatrix*) f.extmin();
    }
    if(distance != NULL)
        *distance = solution->lower_bound();
    if(estimated)
        QMessageBox(QMessageBox::NoIcon,"Warning",
                    QString("Maximum heap size of N=%1 was exceeded. This is an estimated solution, not precise!").arg(maxHeap)).exec();
    return solution->get_solution();
}
double TSP::free_RAM_Bytes(){
    MEMORYSTATUSEX statex;
    statex.dwLength = sizeof (statex); // I misunderstand that
    GlobalMemoryStatusEx (&statex);
    return (double) statex.ullAvailPhys/(1024*1024*1024)*1e9;
}
edge* TSP::BruteForce(double* distance){
    // alloc helper data
    double *d = new double[c+1];
    double *dcutoff = new double[c];
    double total_cutoff = 0;
    int *ii = new int[c+1];
    int *imin = new int[c+1];
    double* co =  new double[c+1];
    int ci = 0;
    //initiate upper bound with primitive solution
    TspMatrix m(data,c);
    for(int j = 0; j < c; j++){
        dcutoff[j] = m.row_min(j);
        total_cutoff += dcutoff[j];
    }
    m.primitive_solution();
    edge* e = sort(m.get_solution(),0);
    double upper_bound = m.lower_bound();
    // initiate pother  helper
    for( ci = 0; ci < c+1; ci++){
        d[ci] = 0;
        ii[ci] = 0;
        imin[ci] = e->from;
        e = e->fw();
    }
    // find solution through trial an error
    ii[0] = ii[c] = 0;                          // solution starts with node 0 allways
    d[0] = 0;
    co[0] = total_cutoff - dcutoff[0];
    ci = 1;
    // now execute recursion
    while ( ci > 0 ){
        // check if all customers are already reached
        if(ci < c){
            // salesman is still on his way
            // is there any other customer left at this point?
            if(ii[ci] == c){
                // no other customers left, go back to the last customer
                // and choose the next possible way
                ii[ci] = 0;
                ci--;
                ii[ci]++;
            }
            else{
                // new trial possible, check that current customer was not
                // satisfied before
                bool satisfied = false;
                for(int k = 0; k < ci; k++)
                    if (ii[k] == ii[ci])
                        satisfied = true;
                if(satisfied){
                    // this customer has already been satisfied, go to next one
                    ii[ci]++;
                }
                else{
                    d[ci] = d[ci-1] + data[ii[ci-1]][ii[ci]];  // new distance
                    co[ci] = co[ci-1] - dcutoff[ii[ci]];
                    if(d[ci] > upper_bound - co[ci])
                        // this trial will be longer than the shortest
                        // known at the moment, try another way
                        ii[ci]++;
                    else
                        // proceed along the current way
                        ci++;
                }
            }

        }
        else{
            // salesman has reached the last customer
            d[ci] = d[ci-1] + data[ii[ci-1]][0]; // add way back to start (node 0)
            if(d[ci] < upper_bound){
                // found a new shortes trial
                for(int k = 0; k < c+1; k++)   imin[k] = ii[k];
                upper_bound = d[ci];
            }
            ci--;
            ii[ci]++;
        }
    }
    if(distance != NULL)
        *distance = upper_bound;
    // generate solution
    edge* solution = new edge;
    solution->from = 0;
    solution->to = imin[1];
    edge* tprev = solution, *t;
    for(int i = 1; i <c; i++){
        t = new edge;
        t->from = imin[i];
        t->to = imin[i+1];
        t->link_suc(tprev);
        tprev = t;
    }
    solution->link_suc(t);
    return solution;
}
edge* TSP::NearestNeighbour(double* distance){
    TspMatrix *T = new TspMatrix(data,c);
    T->primitive_solution();
    if(distance != NULL)
        *distance = T->lower_bound();
    return T->get_solution();
}
edge* TSP::sort(edge* src, int from){
    edge* result = src->fw();
    while(result != NULL && result->from != from && result != src)
        result = result->fw();

    if(result == src && result->from != from)
        return NULL;
    else
        return result;
}
double TSP::solutionlength(TspMatrix* T){
    QList<edge*> *sl = T->solutionlist();
    double sum = 0;
    for(int i = 0; i < sl->length(); i++)
        sum += data[sl->at(i)->from][sl->at(i)->to];
    return sum;
}
